#include "./generator-rf.h"
#include <random>
#include <cmath>
#include <stdexcept>
namespace mri {
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::EvaluateApproximateDemodulated(std::vector<double> &mag_sin,
                                        std::vector<double> &mag_cos) {
    std::random_device rd;
    std::mt19937_64 om;
    om.seed(rd());
    double lbound = 0.0, ubound = 1.0;
    std::uniform_real_distribution<double> rand(lbound, ubound);

    mag_sin.resize(total_samples_);
    mag_cos.resize(total_samples_);
    for (unsigned long i = 0; i < total_samples_; ++i) {
      mag_sin[i] = 0.0;
      mag_cos[i] = 0.0;
      double time = timesteps_[i];
      for (unsigned int voxel = 0; voxel < voxel_num_; ++voxel) {
        mag_sin[i] +=
          amplitudes_[voxel]
          * std::sin(phases_[voxel]
                     )
          * ApproxExp(1,-time/
                     (T2s_decays_norm_[voxel]*T2s_scale_
                      )
                     )
          + (rand(om)-0.5)*noise_amplitude_

          ;
        mag_cos[i] +=
          amplitudes_[voxel]
          * std::cos(phases_[voxel]
                     )
          * ApproxExp(1,-time/
                     (T2s_decays_norm_[voxel]*T2s_scale_
                      )
                     )
          + (rand(om)-0.5)*noise_amplitude_
          
          ;
       }
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::EvaluateDemodulated(std::vector<double> &mag_sin,
                                        std::vector<double> &mag_cos) {
    std::random_device rd;
    std::mt19937_64 om;
    om.seed(rd());
    double lbound = 0.0, ubound = 1.0;
    std::uniform_real_distribution<double> rand(lbound, ubound);

    mag_sin.resize(total_samples_);
    mag_cos.resize(total_samples_);
    for (unsigned long i = 0; i < total_samples_; ++i) {
      mag_sin[i] = 0.0;
      mag_cos[i] = 0.0;
      double time = timesteps_[i];
      for (unsigned int voxel = 0; voxel < voxel_num_; ++voxel) {
        mag_sin[i] +=
          amplitudes_[voxel]
          * std::sin(phases_[voxel]
                     )
          * std::exp(-time/
                     (T2s_decays_norm_[voxel]*T2s_scale_
                      )
                     )
          + (rand(om)-0.5)*noise_amplitude_

          ;
        mag_cos[i] +=
          amplitudes_[voxel]
          * std::cos(phases_[voxel]
                     )
          * std::exp(-time/
                     (T2s_decays_norm_[voxel]*T2s_scale_
                      )
                     )
          + (rand(om)-0.5)*noise_amplitude_
          
          ;
       }
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::EvaluateSignal(std::vector<double> &signal_rf) {
    signal_rf.resize(total_samples_);
    for (unsigned long i = 0; i < total_samples_; ++i) {
      signal_rf[i] = 0.0;
      double time = timesteps_[i];
      for (unsigned int voxel = 0; voxel < voxel_num_; ++voxel) {
        signal_rf[i] +=
          amplitudes_[voxel]
          * std::sin(
                     omega_*time + phases_[voxel]
                     )
          * std::exp(-time/
                     (T2s_decays_norm_[voxel]*T2s_scale_
                      )
                     )
          // ) + ( 0.0 
          //+ np.random.rand()/noise_ratio
          
          ;
      }
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::RandomizeVoxels(){
    std::random_device rd;
    std::mt19937_64 om;
    om.seed(rd());
    double lbound = 0.0, ubound = 1.0;
    std::uniform_real_distribution<double> rand(lbound, ubound);
    amplitudes_.clear();
    T2s_decays_norm_.clear();
    amplitudes_.resize(voxel_num_);
    T2s_decays_norm_.resize(voxel_num_);
    for (double &amp:amplitudes_){
      amp = rand(om);
    }
    for (double &decay:T2s_decays_norm_){
      decay = rand(om);
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::SetAmplitudes(std::vector<double> amplitudes){
    if (amplitudes.size() != voxel_num_)
      throw std::invalid_argument
        ("Number of amplitudes do not fit number of voxels");
    amplitudes_ = amplitudes;
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::SetBase(const std::vector<double> &mag_sin,
                            const std::vector<double> &mag_cos){
    if ((mag_sin.size() != mag_cos.size())
      ||
        (mag_sin.size() != total_samples_)) {
      throw std::invalid_argument
        ("Base signal size does not fit settings of the fitness signal.");
    }
    base_mag_sin_.resize(total_samples_);
    base_mag_cos_.resize(total_samples_);
    for (unsigned long i = 0; i<total_samples_; ++i) {
      base_mag_sin_[i] = mag_sin[i];
      base_mag_cos_[i] = mag_cos[i];
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::SetT2sDecaysNorm(std::vector<double> T2s_decays_norm){
    T2s_decays_norm_ = T2s_decays_norm;
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  // amplitudes -> values[:size/2]
  // T2s_decays_norm -> values[size/2:]
  void GeneratorRF::SetVoxels(const std::vector<double> &values) {
    for (unsigned int i = 0; i < voxel_num_; ++i) {
      amplitudes_[i] = values[i];
      T2s_decays_norm_[i] = values[i+voxel_num_];
    }
  }; 
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  // amplitudes -> values[:size/2]
  // T2s_decays_norm -> values[size/2:]
  void GeneratorRF::GetVoxels(std::vector<double> &values) {
    values.resize(2 * voxel_num_);
    for (unsigned int i = 0; i < voxel_num_; ++i) {
      values[i] = amplitudes_[i];
      values[i+voxel_num_] = T2s_decays_norm_[i];
    }
  }; 
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  double GeneratorRF::ApproxExp(int terms, double x){
    if (terms % 2  == 0) 
      throw std::invalid_argument
        ("Exponential approximation is only available for even terms!");
    
    double result = 1.0;
    for (int i = 1; i < terms+1; ++i) {
      result += 1.0*std::pow(x,i)/std::tgamma(i+1);
      //      std::cout << " %% " << std::tgamma(i+1)<< " "<< i <<" -> "<<result;
    }
    if (result < 0) return 0.0;
    return result;
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  double GeneratorRF::TryApproximateFit(const std::vector<double> &values) {
    SetVoxels(values);
    double diff = 0.0;
    std::vector<double> mag_sin, mag_cos;
    EvaluateApproximateDemodulated(mag_sin, mag_cos);
    
    for (unsigned long i = 0; i < mag_sin.size(); ++i) {
      diff += std::abs((base_mag_sin_[i] - mag_sin[i]));
      diff += std::abs((base_mag_cos_[i] - mag_cos[i]));

      // diff += std::abs((base_mag_sin_[i] - mag_sin[i])
      //                  / base_mag_sin_[i]);
      // diff += std::abs((base_mag_cos_[i] - mag_cos[i])
      //                  / base_mag_cos_[i]);
      // diff += std::abs(
      //                  (base_mag_sin_[i]/base_mag_cos_[i] - mag_sin[i]/mag_cos[i])
      //                  );
    }
    return diff;
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  double GeneratorRF::TryFit(const std::vector<double> &values) {
    SetVoxels(values);
    std::vector<double> mag_sin, mag_cos;
    EvaluateDemodulated(mag_sin, mag_cos);
    double diff = 0.0;
    for (unsigned long i = 0; i < mag_sin.size(); ++i) {
      // diff += std::abs((base_mag_sin_[i] - mag_sin[i]));
      // diff += std::abs((base_mag_cos_[i] - mag_cos[i]));

      diff += std::abs((base_mag_sin_[i] - mag_sin[i])
                       / //(pow2(
                              base_mag_sin_[i]
                         //      )+
                         // pow2(base_mag_cos_[i])
                         // )
                       );
      diff += std::abs((base_mag_cos_[i] - mag_cos[i])
                       / //(pow2(base_mag_sin_[i])+
                         //pow2(
                       base_mag_cos_[i]
                       // )
                       //   )
                       );
      diff += std::abs(
                       (base_mag_sin_[i]/base_mag_cos_[i] - mag_sin[i]/mag_cos[i])
                       // /((base_mag_sin_[i])+
                       //   (base_mag_cos_[i])
                       //   )                       
                       );
    }
    return diff;
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  // period is set in milliseconds (ms)
  void GeneratorRF::InitTimer(unsigned long total_periods, double period_ms,
                             int samples_per_period) {
    total_periods_ = total_periods;
    period_ms_ = period_ms;
    samples_per_period_ = samples_per_period;
    omega_ = 2.0*PI_/period_ms_;
    total_samples_ = total_periods * samples_per_period;
    timesteps_.resize(total_samples_);
    double a_timestep = period_ms/samples_per_period;
    for (unsigned long i = 0; i < total_samples_; ++i) {
      timesteps_[i] = i*a_timestep;      
    }
    // for (double step: timesteps_) std::cout << step<<' ';
    // std::cout<<std::endl;
  };

  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::InitApproximation() {
    exp_terms_.resize(voxel_num_);
    for (auto &term:exp_terms_) {
      term = 1;
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //
  void GeneratorRF::InitVoxels(unsigned int voxel_num, double phase_init,
                              double phase_range, double T2s_scale) {
    T2s_scale_ = T2s_scale;
    if (voxel_num < 2)
      throw std::invalid_argument("We need at least two voxels for phase gradient");    
    voxel_num_ = voxel_num;
    amplitudes_.resize(voxel_num_);
    T2s_decays_norm_.resize(voxel_num_);
    phase_init_ = phase_init;
    phase_range_ = phase_range;
    double phase_step = phase_range_/(voxel_num_ - 1);
    phases_.resize(voxel_num);
    for (unsigned int i =0; i< voxel_num_; ++i) {
      phases_[i] = phase_init_ + i*phase_step;
      //phases_[i] = phase_init_ + std::sin(1.0*i/voxel_num_);
      std::cout<<phases_[i]<<std::endl;
    }
  };
  // ********************************************************************** //
  // ********************************************************************** //
  // ********************************************************************** //

}  // end of namespace mri
